import asyncio
from argparse import Namespace
from concurrent.futures import ThreadPoolExecutor
from contextlib import asynccontextmanager
from typing import ClassVar

from fastapi import FastAPI

from api import config
from evolution import GeneticAlgorithm
from models import Llama, LLMModel
from task import SentimentAnalysis


class MultiProcessOptimizer:
    _instance: ClassVar["MultiProcessOptimizer"] = None
    _running: bool = False
    model_exec: ThreadPoolExecutor = None
    _evolution_model: LLMModel | None = None
    _evaluation_model: LLMModel | None = None

    def __new__(cls, *args, **kwargs):
        # only allow to create one instance (singleton pattern)
        if cls._instance is None:
            cls._instance = super(MultiProcessOptimizer, cls).__new__(cls)
        return cls._instance

    def __init__(
        self, *, event_loop: asyncio.BaseEventLoop, debug: bool = False
    ) -> None:
        self.event_loop = event_loop
        self.debug = debug

    def __enter__(self):
        # TODO allow to customize optimizer

        # initialize worker processes; only 1 worker since prediction is memory-intensive
        # since we only have 1 worker we just save the state in the global namespace which the single worker accesses
        self.model_exec = ThreadPoolExecutor(
            max_workers=1, initializer=self._setup_models
        )
        # make sure that initializer is called
        evolution_model = lambda *args, **kwargs: self.model_exec.submit(
            self._call_evolution_model, *args, **kwargs
        ).result()
        evaluation_model = lambda *args, **kwargs: self.model_exec.submit(
            self._call_evaluation_model, *args, **kwargs
        ).result()

        # currently fix task
        options = Namespace(use_grammar=False, debug=self.debug)
        task = SentimentAnalysis(evaluation_model, options)

        optimizer_class = GeneticAlgorithm
        # optimizer_class = DifferentialEvolution

        self.optimizer = optimizer_class(
            population_size=10,
            task=task,
            evolution_model=evolution_model,
            evaluation_model=evaluation_model,
            run_options=options.__dict__,
        )

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.model_exec.submit(self._release_models)
        self.model_exec.shutdown(False)
        self.model_exec = None

    def _setup_models(self) -> None:
        print("setup models")
        if self._evolution_model is not None:
            raise Exception("Evolution model has already been initialized.")

        # currently fix model
        options = Namespace(
            llama_path="./models/llama-2-13b-chat.Q5_K_M.gguf", chat=True
        )
        self._evolution_model = Llama(options)

        if self._evaluation_model is not None:
            raise Exception("Evaluation model has already been initialized.")

        # currently fix model
        self._evaluation_model = self._evolution_model

    def _release_models(self) -> None:
        del self._evolution_model
        del self._evaluation_model

    def _call_evolution_model(self, *args, **kwargs):
        return self._evolution_model(*args, **kwargs)

    def _call_evaluation_model(self, *args, **kwargs):
        return self._evaluation_model(*args, **kwargs)

    def run_optimization(self, num_iterations: int):
        self._running = True
        self.optimizer.run(num_iterations, debug=self.debug)
        self._running = False

    def get_progress(self):
        if self.optimizer.iterations_pbar is not None:
            return {
                "current": self.optimizer.iterations_pbar.n,
                "total": self.optimizer.iterations_pbar.total,
            }

    def get_run_name(self):
        if self.optimizer.run_directory is not None:
            return self.optimizer.run_directory.name

    async def evolve(self, prompt1: str, prompt2: str):
        return await self.event_loop.run_in_executor(
            self.model_exec, self.optimizer.evolve, prompt1, prompt2
        )


instance: MultiProcessOptimizer | None = None


@asynccontextmanager
async def lifespan(app: FastAPI):
    global instance
    # Load the instance (which runs models in a separate thread)
    instance = MultiProcessOptimizer(
        event_loop=asyncio.get_running_loop(),
        debug=config.DEBUG,
    )
    with instance:
        yield
