import json
import logging
from abc import abstractmethod
from itertools import zip_longest
from pathlib import Path
from typing import Any, Optional

from tqdm import trange

from evoprompt.models import LLMModel
from evoprompt.opt_types import ModelUsage, OptTypeEncoder, Prompt
from evoprompt.task import Task
from evoprompt.utils import initialize_run_directory, log_calls

logger = logging.getLogger(__name__)


PARAPHRASE_PROMPT = """You are given an instruction that describes a task. Write a response that paraphrases the instruction. Only output the paraphrased instruction bracketed in <prompt> and </prompt>."""


@log_calls("Paraphrasing prompts")
def paraphrase_prompts(
    model: LLMModel,
    prompt: str,
    n: int,
    unique_paraphrases: bool = False,
    max_tries: int = 10,
    return_only_unique_paraphrases: bool = False,
) -> tuple[list, ModelUsage]:
    total_usage = ModelUsage()
    paraphrases = []
    num_tries = 0
    while len(paraphrases) < n:
        if num_tries >= max_tries:
            break
        num_tries += 1
        paraphrase, usage = model(
            system_message=PARAPHRASE_PROMPT,
            prompt=prompt,
            prompt_prefix=' Instruction: "',
            prompt_suffix='"',
        )
        total_usage += usage
        if "<prompt>" in paraphrase:
            paraphrase = paraphrase.split("<prompt>")[1].split("</prompt>")[0]
        if not unique_paraphrases or paraphrase not in paraphrases:
            # add paraphrase only if not already present if unique_paraphrases==True
            paraphrases.append(paraphrase)
    if return_only_unique_paraphrases:
        paraphrases = list(set(paraphrases))
    return paraphrases, total_usage


class PromptOptimization:
    total_evaluation_usage: ModelUsage | None = None
    total_evolution_usage: ModelUsage | None = None
    run_directory: Path | None = None
    # P contains the list of prompts at each generation
    P: list[list[Prompt]] | None = None
    # family_tree contains the relation of prompts to its parents
    family_tree: dict[str, tuple[str, ...] | None] | None = None
    # all_prompts contains a list of Prompt objects that took part in the optimization
    all_prompts: dict[str, Prompt] | None = None
    iterations_pbar: Optional[trange] = None

    def __init__(
        self,
        *,
        task: Task,
        evolution_model: LLMModel,
        evaluation_model: LLMModel,
        run_options: dict[str, Any] = {},
    ) -> None:
        self.task = task
        self.evolution_model = evolution_model
        self.evaluation_model = evaluation_model
        self.run_options = run_options

    def evaluate_prompt(self, prompt: str, parents: tuple[Prompt] | None = None):
        parent_histories = (
            [parent.evaluation_history for parent in parents]
            if parents is not None
            else None
        )
        return self.task.evaluate_validation(prompt, parent_histories)

    def add_prompt(
        self,
        prompt: str,
        parents: tuple[Prompt] | None = None,
        meta: dict | None = None,
    ) -> Prompt:
        score, usage, history = self.evaluate_prompt(prompt, parents)
        prompt_object = Prompt(
            content=prompt,
            score=score,
            meta=meta if meta is not None else {},
            usage=usage,
            evaluation_history=history,
        )
        logger.info(
            "Prompt '%s' scores %.2f %s.",
            prompt.replace("\r", "\\r").replace("\n", "\\n"),
            score,
            self.task.metric_name,
        )

        # keep track of prompt
        self.all_prompts[prompt_object.id] = prompt_object
        self.family_tree[prompt_object.id] = (
            tuple(p.id for p in parents) if parents is not None else None
        )
        self.save_snapshot()

        return prompt_object

    def add_prompts(
        self,
        prompts: list[str],
        parents: list[tuple[Prompt]] = iter(()),
        metas: list[dict] = iter(()),
    ):
        return [
            self.add_prompt(prompt, _parents, meta)
            for prompt, _parents, meta in zip_longest(prompts, parents, metas)
        ]

    def get_prompt(self, prompt_id: str):
        return self.all_prompts[prompt_id]

    def get_prompts(self, prompt_ids: list[str]):
        return [self.get_prompt(p_id) for p_id in prompt_ids]

    @abstractmethod
    def save_snapshot(self): ...

    def init_run(self, num_initial_prompts: int, num_iterations: int):
        # family_tree contains the relation of prompts to its parents
        self.family_tree: dict[str, tuple[str, ...] | None] = {}
        # all_prompts contains a list of Prompt objects that took part in the optimization
        self.all_prompts: dict[str, Prompt] = {}
        self.P = []
        self.total_evaluation_usage = ModelUsage()
        self.total_evolution_usage = ModelUsage()
        self.iterations_pbar = trange(
            1, num_iterations + 1, desc="iterations", leave=True
        )

        self.run_directory = initialize_run_directory(self.evolution_model)
        self.save_snapshot()

        paraphrases, paraphrase_usage = paraphrase_prompts(
            self.evolution_model, self.task.base_prompt, n=num_initial_prompts - 1
        )
        self.total_evolution_usage += paraphrase_usage
        logger.info("Paraphrased prompt '%s': %s.", self.task.base_prompt.replace("\r", "\\r").replace("\n", "\\n"), paraphrases)

        # the initial prompts
        initial_prompts = [self.task.base_prompt] + paraphrases
        initial_prompts = self.add_prompts(
            initial_prompts, metas=[{"gen": 0} for _ in initial_prompts]
        )
        # - Initial prompts P0 = {p1, p2, . . . , pN }
        self.P.append(initial_prompts)

        # accumulate usage
        for prompt in initial_prompts:
            self.total_evaluation_usage += prompt.usage
        self.save_snapshot()

    def save_snapshot(self):
        with open(self.run_directory / "snapshot.json", "w") as f:
            json.dump(
                {
                    "all_prompts": self.all_prompts,
                    "family_tree": self.family_tree,
                    "P": [
                        [prompt.id for prompt in population] for population in self.P
                    ],
                    "T": self.family_tree,
                    "N": self.iterations_pbar.n,
                    "task": {
                        "name": self.task.__class__.__name__,
                        "validation_dataset": self.task.validation_dataset.info.dataset_name,
                        "test_dataset": self.task.test_dataset.info.dataset_name,
                        "metric": self.task.metric_name,
                        "use_grammar": self.task.use_grammar,
                    },
                    "model": {"name": self.evolution_model.__class__.__name__},
                    "evaluation_usage": self.total_evaluation_usage,
                    "evolution_usage": self.total_evolution_usage,
                    "run_options": self.run_options,
                },
                f,
                indent=4,
                cls=OptTypeEncoder,
            )


def load_snapshot(path: Path):
    import json

    with path.open("r") as f:
        snapshot = json.load(f)
    return (
        snapshot["family_tree"],
        snapshot["P"],
        snapshot["S"],
        snapshot["T"],
        snapshot["N"],
    )
